
#include "time_util.h"
#include "UCT.h"
#include "SFMT.h"
#include <math.h>

#define EXPAND_TRIALS 100
#define MAX_DEPTH 1500

uint64_t rand_cache;
int rand_ix = 64;

int randomDir()
{
    if (rand_ix > 62)
    {
        rand_cache = gen_rand64();
        rand_ix = 0;
    }
    int next_dir = (rand_cache >> rand_ix) & 0x3;
    rand_ix += 2;
    return next_dir;
}

void Node::set(Map& position)
{
    hash = position.hash;
    aged = false;
    visits = 1;
	int moving = position.moving;
    for (int i=0; i < 4; i++)
    {
        position.doStep(i);
        int* bike = position.bikes[moving];
        viable[i] = !position.isWall(bike);
        children[i] = position.hash;
        wins[i] = 1;
        trials[i] = 1;
        position.undoStep(i);
    }
}

NodeTable::NodeTable() : table(NULL)
{
    resizeTable(1000000);
}

void NodeTable::resizeTable(int size)
{
    if (table != NULL)
    {
        delete[] table;
    }
    table = new NodeTableCell[size];
    table_size = size;
}

void NodeTable::get(Node& result, Map& position)
{
    NodeTableCell* cell = &table[position.hash % table_size];
    if (cell->high_count.hash == position.hash)
    {
        cell->high_count.aged = false;
        result = cell->high_count;
    }
    else if (cell->always.hash == position.hash)
    {
        cell->always.aged = false;
        result = cell->always;
    } else {
        result.set(position);
    }
}

void NodeTable::set(const Node& node)
{
    NodeTableCell* cell = &table[node.hash % table_size];
    if (cell->high_count.aged || cell->high_count.visits < node.visits)
    {
        cell->high_count = node;
    } else {
        cell->always = node;
    }
}

void NodeTable::ageTable()
{
    for (int i=0; i < table_size; i++)
    {
        table[i].high_count.aged = true;
    }
}

UCTSearch::UCTSearch() : steps_taken(0), num_playouts(0),
    total_depth(0), depth(0)
{
#ifdef TRACE_TREE
    print_tree = false;
#endif
    nodes = new NodeTable();
}

int UCTSearch::lightStep(const Map& position)
{
    int direction = 0;
    int best_dir = 0;
    int checked = 0;
    int x, y;
    do {
        do {
            direction = randomDir();
        } while ((checked & (1 << direction)) != 0);
        checked |= 1 << direction;
        x = position.bikes[position.moving][0];
        y = position.bikes[position.moving][1];
        switch (direction)
        {
            case 0:
                y -= 1;
                break;
            case 1:
                x += 1;
                break;
            case 2:
                y += 1;
                break;
            case 3:
                x -= 1;
                break;
        }
        // only stop first player from stepping on second
        // second bike will actually be a wall when move executes
        if (!(position.isWall(x, y) || (position.moving == 0
                    && position.bikes[1][0] == x
                    && position.bikes[1][1] == y)))
        {
            best_dir = direction;
            break;
        }
    } while (checked != 0xF);
    return best_dir;
}

double UCTSearch::playout(const Map& start)
{
    Map position = start;
    while (!position.isEnd())
    {
        int direction = lightStep(position);
        position.doPlayoutStep(direction);
        ++steps_taken;
    }
    return position.score();
}

#ifdef TRACE_TREE
const char* bstr(bool b)
{
    return b ? "true" : "false";
}
#endif

double UCTSearch::searchTree(Map& position)
{
    ++depth;
	if (position.isEnd())
	{
		return position.score();
	}
    Node root;
    nodes->get(root, position);
#ifdef CHECK_HASH
    for (int i=0; i < 4; i++)
    {
        position.doStep(i);
        uint64_t child_hash = position.hash;
        position.undoStep(i);
        if (child_hash != root.children[i])
        {
            fprintf(stderr, "\n\n\n\n\n\n\n\nHash collision found\n");
            fprintf(stderr, "parent %llX child (%d) %llX expected %llX\n",
                    position.hash, i, child_hash, root.children[i]);
            exit(1);
            root.set(position);
            break;
        }
    }
#endif
    int best_dir = 0;
    double best_score = -1;
    double log_tries = 2 * log((double)root.visits);
    for (int i=0; i < 4; i++)
    {
        if (!root.viable[i])
            continue;
        double score = (root.wins[i] / root.trials[i])
            + sqrt(log_tries / root.trials[i]);
        if (score > best_score)
        {
            best_score = score;
            best_dir = i;
        }
    }
#ifdef TRACE_TREE
    if (print_tree)
    {
        std::string board_str;
        position.toString(board_str);
        fprintf(stderr, "board:\n%s", board_str.c_str());
        fprintf(stderr, "depth %d visits %lld side %d pos %d,%d best_dir %d win_per %.2f\n",
                depth, root.visits, position.moving,
                position.bikes[position.moving][0],
                position.bikes[position.moving][1], best_dir,
                (root.wins[best_dir] / (double)root.trials[best_dir]) * 100.);
        fprintf(stderr, "viable %8s %8s %8s %8s\n",
                bstr(root.viable[0]), bstr(root.viable[1]),
                bstr(root.viable[2]), bstr(root.viable[3]));
        fprintf(stderr, "wins   %8.1f %8.1f %8.1f %8.1f\n",
                root.wins[0], root.wins[1], root.wins[2], root.wins[3]);
        fprintf(stderr, "trials %8.0f %8.0f %8.0f %8.0f\n",
                root.trials[0], root.trials[1], root.trials[2], root.trials[3]);
        fprintf(stderr, "\n");
    }
#endif
#ifdef CHECK_DEPTH
    if (depth >= MAX_DEPTH)
    {
        std::string board_str;
        position.toString(board_str);
        fprintf(stderr, "Hit depth of %d with position:\n", MAX_DEPTH);
        fprintf(stderr, "%s", board_str.c_str());
    }
#endif
    double new_win = 0;
    if (!root.viable[best_dir])
    {
        new_win = playout(position);
        ++num_playouts;
        if (position.moving == 1)
            new_win = 1 - new_win;
    }
	else if (root.trials[best_dir] >= EXPAND_TRIALS && depth < MAX_DEPTH)
    {
        position.doStep(best_dir);
        new_win = 1 - searchTree(position);
        position.undoStep(best_dir);
    } else {
        position.doStep(best_dir);
        new_win += playout(position);
        position.undoStep(best_dir);
        if (position.moving == 1)
            new_win = 1 - new_win;
        ++num_playouts;
    }
    root.visits += 1;
    root.wins[best_dir] += new_win;
    root.trials[best_dir] += 1;
    nodes->set(root);
    return new_win;
}

int UCTSearch::search(Map& position, uint64_t start_time, int search_time)
{
    depth = 0;
    max_depth = 0;
    total_depth = 0;
    num_playouts = 0;
	uint64_t num_searches = 0;
    steps_taken = 0;
    int best_dir = 0;
    Node root;
    nodes->get(root, position);
	uint64_t initial_visits = root.visits;
    int num_viable = 0;
    for (int i=0; i < 4; i++)
    {
        if (root.viable[i])
        {
            ++num_viable;
            best_dir = i;
        }
    }
    if (num_viable < 1)
        return best_dir;
    nodes->set(root);
    uint64_t next_check = 1000;
    while (true)
    {
        searchTree(position);
		++num_searches;
        total_depth += depth;
        if (depth > max_depth)
        {
            max_depth = depth;
#ifdef CHECK_DEPTH
            if (depth >= MAX_DEPTH)
            {
                std::string board_str;
                position.toString(board_str);
                fprintf(stderr, "from position:\n");
                fprintf(stderr, "%s", board_str.c_str());
                exit(1);
            }
#endif
        }
        depth = 0;
        if (num_searches >= next_check)
        {
            uint64_t now = millis();
            int length = now - start_time;
            if (length >= search_time)
                break;
            int time_left = search_time - length;
            length = length > 5 ? length : 5;
            double spm = num_searches / (double)length;
            if (time_left > 50)
            {
                next_check = (spm * (time_left / 4)) + num_searches;
            }
            else if (time_left > 10)
            {
                next_check = num_searches + spm;
            } else {
                next_check = num_searches + 1;
            }
        }
    }
    nodes->get(root, position);
    double best_score = 0;
    for (int i=0; i < 4; i++)
    {
        if (root.viable[i] && root.wins[i] > best_score)
        {
            best_score = root.wins[i];
            best_dir = i;
        }
    }
#ifdef SEARCH_REPORT
    double best_per = (root.wins[best_dir] / root.trials[best_dir]) * 100.0;
    fprintf(stderr, "\ntime %lld searches %lld playouts %lld win %.1f%% avg_len %.1f\t\t\n",
            millis() - start_time,
            num_searches, num_playouts, best_per,
            (steps_taken / 2.0) / (num_playouts ? num_playouts : 1));
    fprintf(stderr, "depth avg %lld max %d\t\t\n", total_depth/num_searches,
            max_depth);
    fprintf(stderr, "Starting visits %d\t\n", initial_visits);
    fprintf(stderr, "wins   %8.1f %8.1f %8.1f %8.1f\t\t\n",
            root.wins[0], root.wins[1], root.wins[2], root.wins[3]);
    fprintf(stderr, "trials %8.1f %8.1f %8.1f %8.1f\t\t\n",
            root.trials[0], root.trials[1], root.trials[2], root.trials[3]);
#endif
    return best_dir;
}

